﻿#include <time.h>
#include <android/log.h>
#include "com_example_mosaicize_MosaicizeActivity.h"

#define min(a, b) (\
		   (a < b) ? a : b \
		   )

struct Colors_t{
    unsigned long blue;
    unsigned long green;
    unsigned long red;
    unsigned long alpha;
};

struct Pixel_t {
    unsigned char blue;
    unsigned char green;
    unsigned char red;
    unsigned char alpha;
};

JNIEXPORT void JNICALL Java_com_example_mosaicize_MosaicizeActivity_getMosaicizedPixels2
(JNIEnv *env, jobject thiz, jintArray array, jint width, jint height, jint size, jobject version){

    // 引数が変なときはそのまま返る
    if (height == 0 || width == 0 || size == 0) {
	return;
    }

    // pixel を C の配列として取得
    jboolean isCopy;
    jint *pixels = (*env)->GetIntArrayElements(env, array, &isCopy);

    // enum オブジェクトを比較するために取得
    jclass activity_class = (*env)->FindClass(env, "com/example/mosaicize/MOSAIC_PROCESS");
    jobject enum_C_obj, enum_ASM_obj, enum_NEON_obj;

    if (activity_class){
	jfieldID enum_C = (*env)->GetStaticFieldID(env, activity_class, "C", "Lcom/example/mosaicize/MOSAIC_PROCESS;");
	if (enum_C){
	    enum_C_obj = (*env)->GetStaticObjectField(env, activity_class, enum_C);
	}
	
	jfieldID enum_ASM = (*env)->GetStaticFieldID(env, activity_class, "ASM", "Lcom/example/mosaicize/MOSAIC_PROCESS;");
	if (enum_ASM) {
	    enum_ASM_obj = (*env)->GetStaticObjectField(env, activity_class, enum_ASM);
	}

	jfieldID enum_NEON = (*env)->GetStaticFieldID(env, activity_class, "NEON", "Lcom/example/mosaicize/MOSAIC_PROCESS;");
	if(enum_NEON){
	    enum_NEON_obj = (*env)->GetStaticObjectField(env, activity_class, enum_NEON);
	}
    } else {
	__android_log_print(ANDROID_LOG_INFO, "MOSAICIZE", "FindClass failed");
	goto error_exit;
    }

    int x, y;

#ifdef MEASURE
    struct timespec tv1, tv2, tv3, tv4;
    unsigned long time1 = 0, time2 = 0, time3 = 0;
#endif

    for (y = 0; y < height; y += size) {
	for (x = 0; x < width; x += size) {
	    int dx, dy;
	    int count = 0;
	    struct Colors_t colors = {0, 0, 0, 0};
	    struct Pixel_t *ppixel;

#ifdef MEASURE
	    clock_gettime(CLOCK_MONOTONIC, &tv1);
#endif

	    // ブロック毎に計算するときに、写真の横幅や高さを超えないように
	    int x_limit = min((x + size), width);
	    int y_limit = min((y + size), height);

	    if ((*env)->IsSameObject(env, version, enum_C_obj)){
		// C言語バージョン
		// ブロック内のピクセルの要素ごとの和をとる
		for (dy = y; dy < y_limit; dy++) {
		    for (dx = x; dx < x_limit; dx++) {
			int i = dx + dy * width;
			

			ppixel = (struct Pixel_t *)(&pixels[i]);
			colors.alpha += ppixel->alpha;
			colors.red   += ppixel->red;
			colors.green += ppixel->green;
			colors.blue  += ppixel->blue;

			count++;
		    }
		}
	    } else {
		if ((*env)->IsSameObject(env, version, enum_ASM_obj)){
		    // インラインアセンブラバージョン
		    // ブロック内のピクセルの要素ごとの和をとる
		    for (dy = y; dy < y_limit; dy++) {
			for (dx = x; dx < x_limit; dx++) {
			    int i = dx + dy * width;

			    asm volatile ("ldrb r0, [%1];"          /* r0 = pixels[i] & 0xff*/
					  "ldr  r1, [%0];"          /* r1 = colors.blue */
					  "add  r1, r1, r0;"        /* blue += r0 */
					  "str  r1, [%0];"
					  "ldrb r0, [%1, #1];"      /* r0 = pixels[i] >> 8 & 0xff */
					  "ldr  r1, [%0, #4];"      /* r1 = colors.green */
					  "add  r1, r1, r0;"        /* green += r0 */
					  "str  r1, [%0, #4];"
					  "ldrb r0, [%1, #2];"      /* r0 = pixels[i] >> 8 & 0xff */
					  "ldr  r1, [%0, #8];"      /* r1 = colors.red */
					  "add  r1, r1, r0;"        /* red += r0 */
					  "str  r1, [%0, #8];"
					  "ldrb r0, [%1, #3];"      /* r0 = pixels[i] & 0xff*/
					  "ldr  r1, [%0, #12];"     /* r1 = colors.alpha */
					  "add  r1, r1, r0;"        /* alpha += r0 */
					  "str  r1, [%0, #12];"
					  : /* no output */
					  : "r"(&colors), "r"(&pixels[i]) /* input */
					  : "memory", "r0", "r1" /* clobber list */
					  );
			    count++;
			}
		    }
		} else {
		    if ((*env)->IsSameObject(env, version, enum_NEON_obj)){
			// NEONバージョン
			// ブロック内のピクセルの要素ごとの和をとる
			for (dy = y; dy < y_limit; dy++) {
			    // 8ピクセル毎にまず計算
			    for (dx = x; (x_limit - dx)/8 > 0 ; dx+=8) {
				int i = dx + dy * width;
				asm volatile ("vld4.8       {d0-d3}, [%1];" /* 8ピクセル分読み取り*/
					      "vpaddl.u8  q2, q0;"       /* q2: g g g g b b b b 16bit */
					      "vpaddl.u8  q5, q1;"       /* q5: a a a a r r r r 16bit */
					      "vpaddl.u16 q3, q2;"       /* q3: g g b b         32bit */
					      "vpaddl.u16 q6, q5;"       /* q6: a a r r         32bit */
					      "vpaddl.u32 q4, q3;"       /* q4: g b             64bit */
					      "vpaddl.u32 q7, q6;"       /* q7: a r             64bit */
					      "vldm       %0, {q0};"     /* q0: a r g b */
					      "vswp       d9, d14;"      /* d14: 0 g, d9: 0 r*/
					      "vtrn.32    q4, q7;"       /* q4: a r g b */
					      "vadd.i32   q4, q0;"       /* colors の要素との和を求める */
					      "vstm       %0, {q4};"     /* b g r a の順にメモリに保存*/
					      : /* no output */
					      : "r"(&colors), "r"(&pixels[i]) /* input */
					      : "cc", "memory",               /* clobber list */
					      "q0", "q1", "q2", "q3",
					      "q4", "q5", "q6", "q7"
					      );
				count+=8;
			    }
			    // 端数を計算
			    // dx は初期化済み
			    for (; dx < x_limit; dx++) {
				int i = dx + dy * width;

				asm volatile ("ldrb r0, [%1];"          /* r0 = pixels[i] & 0xff*/
					  "ldr  r1, [%0];"          /* r1 = colors.blue */
					  "add  r1, r1, r0;"        /* blue += r0 */
					  "str  r1, [%0];"
					  "ldrb r0, [%1, #1];"      /* r0 = pixels[i] >> 8 & 0xff */
					  "ldr  r1, [%0, #4];"      /* r1 = colors.green */
					  "add  r1, r1, r0;"        /* green += r0 */
					  "str  r1, [%0, #4];"
					  "ldrb r0, [%1, #2];"      /* r0 = pixels[i] >> 8 & 0xff */
					  "ldr  r1, [%0, #8];"      /* r1 = colors.red */
					  "add  r1, r1, r0;"        /* red += r0 */
					  "str  r1, [%0, #8];"
					  "ldrb r0, [%1, #3];"      /* r0 = pixels[i] & 0xff*/
					  "ldr  r1, [%0, #12];"     /* r1 = colors.alpha */
					  "add  r1, r1, r0;"        /* alpha += r0 */
					  "str  r1, [%0, #12];"
					  : /* no output */
					  : "r"(&colors), "r"(&pixels[i]) /* input */
					  : "memory", "r0", "r1" /* clobber list */
					  );
				count++;
			    }
			}
		    } else {
			__android_log_print(ANDROID_LOG_INFO, "MOSAICIZE", "found nothing to do.");
		    }
		}
	    }
	
	    if (0 == count){
		break;
	    }

#ifdef MEASURE
	    clock_gettime(CLOCK_MONOTONIC, &tv2);
#endif

	    // 要素毎に平均をとる
	    colors.alpha /= count;
	    colors.red /= count;
	    colors.green /= count;
	    colors.blue /= count;

#ifdef MEASURE
	    clock_gettime(CLOCK_MONOTONIC, &tv3);
#endif

	    // ブロック内のピクセル全てに平均値を適用する
	    for (dy = y; dy < y_limit; dy++) {
		for (dx = x; dx < x_limit; dx++) {
		    int i = dx + dy * width;
		    ppixel = (struct Pixel_t *)(&pixels[i]);
		    ppixel->alpha = colors.alpha;
		    ppixel->red   = colors.red;
		    ppixel->green = colors.green;
		    ppixel->blue  = colors.blue;
		}
	    }

#ifdef MEASURE
	    clock_gettime(CLOCK_MONOTONIC, &tv4);

	    time1 += (tv2.tv_sec - tv1.tv_sec) * 1000000000 + tv2.tv_nsec - tv1.tv_nsec;
	    time2 += (tv3.tv_sec - tv2.tv_sec) * 1000000000 + tv3.tv_nsec - tv2.tv_nsec;
	    time3 += (tv4.tv_sec - tv3.tv_sec) * 1000000000 + tv4.tv_nsec - tv3.tv_nsec;
#endif
	}
    }
#ifdef MEASURE
    __android_log_print(ANDROID_LOG_INFO, "MOSAICIZE", "span 1 = %d ns", time1);
    __android_log_print(ANDROID_LOG_INFO, "MOSAICIZE", "span 2 = %d ns", time2);
    __android_log_print(ANDROID_LOG_INFO, "MOSAICIZE", "span 3 = %d ns", time3);
#endif

error_exit:
    (*env)->ReleaseIntArrayElements(env, array, pixels, 0);
}
