package com.example.mosaicize;

import android.app.Activity;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Color;
import android.graphics.Bitmap.Config;
import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.Button;
import android.widget.ImageView;
import android.widget.TextView;

enum MOSAIC_PROCESS {
	JAVA, C, ASM, NEON
};

public class MosaicizeActivity extends Activity {
	private Bitmap bitmapOriginal, bitmap;
	private ImageView imageView;
	private boolean isEffected = false;
	protected MOSAIC_PROCESS pattern;
	private int blockSize = 32;

	/** Called when the activity is first created. */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		bitmapOriginal = BitmapFactory.decodeResource(getResources(),
				R.drawable.per);

		if (bitmapOriginal != null) {
			bitmap = bitmapOriginal.copy(Config.ARGB_8888, true);
		}
		setContentView(R.layout.main);

		imageView = (ImageView) findViewById(R.id.imageView1);
		imageView.setImageBitmap(bitmap);

		Button javaButton = (Button) findViewById(R.id.java);

		javaButton.setOnClickListener(new OnClickListener() {
			@Override
			public void onClick(View v) {
				if (isEffected) {
					return;
				}
				mosaicize(MOSAIC_PROCESS.JAVA, blockSize);
				isEffected = true;
				imageView.invalidate(); // onDraw()Ă΂
			}
		});

		Button cButton = (Button) findViewById(R.id.C);

		cButton.setOnClickListener(new OnClickListener() {
			@Override
			public void onClick(View v) {
				if (isEffected) {
					return;
				}
				mosaicize(MOSAIC_PROCESS.C, blockSize);
				isEffected = true;
				imageView.invalidate(); // onDraw()Ă΂
			}
		});

		Button asmButton = (Button) findViewById(R.id.ASM);

		asmButton.setOnClickListener(new OnClickListener() {
			@Override
			public void onClick(View v) {
				if (isEffected) {
					return;
				}
				mosaicize(MOSAIC_PROCESS.ASM, blockSize);
				isEffected = true;
				imageView.invalidate(); // onDraw()Ă΂
			}
		});
		Button neonButton = (Button) findViewById(R.id.NEON);
		neonButton.setOnClickListener(new OnClickListener() {

			@Override
			public void onClick(View v) {
				if (isEffected) {
					return;
				}
				mosaicize(MOSAIC_PROCESS.NEON, blockSize);
				isEffected = true;
				imageView.invalidate(); // onDraw()Ă΂
			}
		});

		Button resetButton = (Button) findViewById(R.id.reset);
		resetButton.setOnClickListener(new OnClickListener() {

			@Override
			public void onClick(View v) {
				isEffected = false;
				if (bitmap != null) {
					bitmap.recycle();
				}
				bitmap = bitmapOriginal.copy(Config.ARGB_8888, true);

				imageView.setImageBitmap(bitmap);
				imageView.invalidate(); // onDraw()Ă΂

				showConsumedTime(0);
			}

		});
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();
		bitmap.recycle();
		bitmapOriginal.recycle();
	}

	private void getMosaicizedPixels(int[] pixels, int width, int height,
			int size) {

		// ςȂƂ͂̂܂ܕԂ
		if (height == 0 || width == 0 || size == 0) {
			return;
		}

		for (int y = 0; y < height; y += size) {
			for (int x = 0; x < width; x += size) {
				int red = 0, green = 0, blue = 0, alpha = 0;

				int count = 0;
				// ubNɌvZƂɁAʐ^̉⍂𒴂Ȃ悤
				int yLimit = Math.min((y + size), height);
				int xLimit = Math.min((x + size), width);

				// ubÑsNZ̗vfƂ̘aƂ
				for (int dy = y; dy < yLimit; dy++) {
					for (int dx = x; dx < xLimit; dx++) {
						int pos = dx + dy * width;

						red += Color.red(pixels[pos]);
						green += Color.green(pixels[pos]);
						blue += Color.blue(pixels[pos]);
						alpha += Color.alpha(pixels[pos]);
						count++;
					}
				}

				if (count == 0) {
					return;
				}

				// vfɕςƂ
				alpha /= count;
				red /= count;
				green /= count;
				blue /= count;

				// ubÑsNZSĂɕϒlKp
				for (int dy = y; dy < yLimit; dy++) {
					for (int dx = x; dx < xLimit; dx++) {
						int pos = dx + dy * width;
						pixels[pos] = Color.argb(alpha, red, green, blue);
					}
				}
			}
		}
	}

	private native void getMosaicizedPixels2(int[] pixels, int width,
			int height, int size, MOSAIC_PROCESS version);

	private void mosaicize(MOSAIC_PROCESS pattern, int tileSize) {
		int height = bitmap.getHeight();
		int width = bitmap.getWidth();
		int pixels[] = new int[width * height];

		bitmap.getPixels(pixels, 0, width, 0, 0, width, height);

		long start = System.currentTimeMillis();

		switch (pattern) {
		case JAVA:
			// java version
			getMosaicizedPixels(pixels, width, height, tileSize);
			break;
		case C:
		case ASM:
		case NEON:
			getMosaicizedPixels2(pixels, width, height, tileSize, pattern);
			break;
		default:
			break;
		}
		long end = System.currentTimeMillis();

		bitmap.setPixels(pixels, 0, width, 0, 0, width, height);
		long consumedTime = end - start;
		Log.i("SHOWBITMAP", "Ԃ́A" + consumedTime + "msł");
		showConsumedTime(consumedTime);
	}

	private void showConsumedTime(long consumedTime) {
		TextView textView = (TextView) findViewById(R.id.textView1);
		textView.setText("Ԃ́A" + consumedTime + "msł");
	}

	static {
		System.loadLibrary("mosaicize");
	}
}
