//
//  MosaicizeiPadAppDelegate.m
//  MosaicizeiPad
//
//  Created by たけべけんいち on 11/07/04.
//  Copyright 2011 __MyCompanyName__. All rights reserved.
//

#import "MosaicizeiPadAppDelegate.h"

@implementation MosaicizeiPadAppDelegate

@synthesize window=_window;

#define min(a, b) ((a < b) ? a : b)

enum MOSAIC_PROCESS {
	C, ASM, NEON
};

struct Colors_t{
    unsigned long red;
    unsigned long green;
    unsigned long blue;
    unsigned long alpha;
};

struct Pixel_t {
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
};

- (UIImage* )getMosaicizedPixels2:(UIImage *)_image Mosaic_Version:(int)version
{
	CGImageRef image_ref = _image.CGImage;
	int width = _image.size.width; 
	int height = _image.size.height;
    
    // 画像を格納するメモリーを確保
	unsigned int *pixels = malloc(sizeof(unsigned int) * height * width);
    
	NSUInteger bytesPerPixel = 4;
	NSUInteger bytesPerRow = bytesPerPixel * width;
	NSUInteger bitsPerComponent = CGImageGetBitsPerComponent(image_ref);
	CGColorSpaceRef colorspace = CGColorSpaceCreateDeviceRGB();
	
	// いったんオフスクリーンを作り、そこに描画
	CGContextRef context = CGBitmapContextCreate(pixels, width, height, 
                                                 bitsPerComponent, bytesPerRow, colorspace,
                                                 kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big);
	CGContextDrawImage(context, CGRectMake(0, 0, width, height), image_ref);
	CGContextRelease(context);
    
    start = [NSDate date];
    //
	// モザイク処理の中身
    //
    int x, y, size = 32;
    
    for (y = 0; y < height; y += size) {
        for (x = 0; x < width; x += size) {
            int dx, dy;
            int count = 0;
            struct Colors_t colors = {0, 0, 0, 0};
            struct Pixel_t *ppixel;
            
            // ブロックごとに計算するときに、写真の横幅や高さを超えないように
            int x_limit = min((x + size), width);
            int y_limit = min((y + size), height);
            
            if (version == C){
                // C言語バージョン
                // ブロック内のピクセルの要素ごとの和をとる
                for (dy = y; dy < y_limit; dy++) {
                    for (dx = x; dx < x_limit; dx++) {
                        int i = dx + dy * width;
                        ppixel = (struct Pixel_t *)(&pixels[i]);
                        colors.alpha += ppixel->alpha;
                        colors.red   += ppixel->red;
                        colors.green += ppixel->green;
                        colors.blue  += ppixel->blue;
                        count++;
                    }
                }
            } else if (version == ASM) {
                    // インラインアセンブラバージョン
                    // ブロック内のピクセルの要素ごとの和をとる
                    for (dy = y; dy < y_limit; dy++) {
                        for (dx = x; dx < x_limit; dx++) {
                            int i = dx + dy * width;
                            
                            asm volatile ("ldrb r0, [%1];"          
                                          "ldr  r1, [%0];"        
                                          "add  r1, r1, r0;"        
                                          "str  r1, [%0];"
                                          "ldrb r0, [%1, #1];"      
                                          "ldr  r1, [%0, #4];"     
                                          "add  r1, r1, r0;"    
                                          "str  r1, [%0, #4];"
                                          "ldrb r0, [%1, #2];"      
                                          "ldr  r1, [%0, #8];" 
                                          "add  r1, r1, r0;"        
                                          "str  r1, [%0, #8];"
                                          "ldrb r0, [%1, #3];"      
                                          "ldr  r1, [%0, #12];"    
                                          "add  r1, r1, r0;"      
                                          "str  r1, [%0, #12];"
                                          : /* no output */
                                          : "r"(&colors), "r"(&pixels[i]) /* input */
                                          : "memory", "r0", "r1" /* clobber list */
                                          );
                            count++;
                        }
                    }
                } else if (version == NEON) {
                        // NEONバージョン
                        // ブロック内のピクセルの要素ごとの和をとる
                        for (dy = y; dy < y_limit; dy++) {
                            // 8ピクセルごとにまず計算
                            for (dx = x; (x_limit - dx)/8 > 0 ; dx+=8) {
                                int i = dx + dy * width;
                                asm volatile ("vld4.8       {d0-d3}, [%1];" /* 8ピクセル分読み取り*/
                                              "vpaddl.u8  q2, q0;"       
                                              "vpaddl.u8  q5, q1;"    
                                              "vpaddl.u16 q3, q2;"       
                                              "vpaddl.u16 q6, q5;"    
                                              "vpaddl.u32 q4, q3;"       
                                              "vpaddl.u32 q7, q6;"       
                                              "vldm       %0, {q0};"   
                                              "vswp       d9, d14;"      
                                              "vtrn.32    q4, q7;"      
                                              "vadd.i32   q4, q0;"       
                                              "vstm       %0, {q4};"     
                                              : /* no output */
                                              : "r"(&colors), "r"(&pixels[i]) /* input */
                                              : "cc", "memory",               /* clobber list */
                                              "q0", "q1", "q2", "q3",
                                              "q4", "q5", "q6", "q7"
                                              );
                                count+=8;
                            }
                            // 端数を計算
                            // dx は初期化済み
                            for (; dx < x_limit; dx++) {
                                int i = dx + dy * width;
                                
                                asm volatile ("ldrb r0, [%1];"          /* r0 = pixels[i] & 0xff*/
                                              "ldr  r1, [%0];"          /* r1 = colors.blue */
                                              "add  r1, r1, r0;"        /* blue += r0 */
                                              "str  r1, [%0];"
                                              "ldrb r0, [%1, #1];"      /* r0 = pixels[i] >> 8 & 0xff */
                                              "ldr  r1, [%0, #4];"      /* r1 = colors.green */
                                              "add  r1, r1, r0;"        /* green += r0 */
                                              "str  r1, [%0, #4];"
                                              "ldrb r0, [%1, #2];"      /* r0 = pixels[i] >> 8 & 0xff */
                                              "ldr  r1, [%0, #8];"      /* r1 = colors.red */
                                              "add  r1, r1, r0;"        /* red += r0 */
                                              "str  r1, [%0, #8];"
                                              "ldrb r0, [%1, #3];"      /* r0 = pixels[i] & 0xff*/
                                              "ldr  r1, [%0, #12];"     /* r1 = colors.alpha */
                                              "add  r1, r1, r0;"        /* alpha += r0 */
                                              "str  r1, [%0, #12];"
                                              : /* no output */
                                              : "r"(&colors), "r"(&pixels[i]) /* input */
                                              : "memory", "r0", "r1" /* clobber list */
                                              );
                                count++;
                            }
                        }
                    } 
            
            if (0 == count){
                break;
            }
            
            // 要素毎に平均をとる
            colors.alpha /= count;
            colors.red /= count;
            colors.green /= count;
            colors.blue /= count;
            
            // ブロック内のピクセル全てに平均値を適用する
            for (dy = y; dy < y_limit; dy++) {
                for (dx = x; dx < x_limit; dx++) {
                    int i = dx + dy * width;
                    ppixel = (struct Pixel_t *)(&pixels[i]);
                    ppixel->alpha = colors.alpha;
                    ppixel->red   = colors.red;
                    ppixel->green = colors.green;
                    ppixel->blue  = colors.blue;
                }
            }
        }
    }
     
    interval = [[NSDate date] timeIntervalSinceDate:start];
    
    // モザイク処理された画像データからUIImageを作る
	CGContextRef mosaicized_context = CGBitmapContextCreate(pixels, width, height, bitsPerComponent, bytesPerRow, colorspace, 
                                    kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big); 
    CGImageRef mosaicized_image_ref = CGBitmapContextCreateImage(mosaicized_context);
	mosaicized_image = [UIImage imageWithCGImage:mosaicized_image_ref];
	CGContextRelease(mosaicized_context);
    CGImageRelease(mosaicized_image_ref);
	CGColorSpaceRelease(colorspace);
    
    free(pixels); 
	
    NSString *str = [[NSString alloc] initWithFormat:@"かかった時間は、%4.0fmsです", interval*1000];
    label.text = str;
    [str release];
    
	return mosaicized_image;
}


- (void)c_buttonPressed
{
    imageview.image = [self getMosaicizedPixels2:image Mosaic_Version:C];
}

- (void)asm_buttonPressed
{
    imageview.image = [self getMosaicizedPixels2:image Mosaic_Version:ASM];
}

- (void)neon_buttonPressed
{
    imageview.image = [self getMosaicizedPixels2:image Mosaic_Version:NEON];
}

- (void)reset_buttonPressed
{
    [image release];
    NSString *path;
    path = [[NSBundle mainBundle] pathForResource:@"cat" ofType:@"jpg"];
    image = [[UIImage alloc] initWithContentsOfFile:path];
    
    imageview.image = image;
    label.text = @"";
}

- (BOOL)application:(UIApplication *)application didFinishLaunchingWithOptions:(NSDictionary *)launchOptions
{
    // Override point for customization after application launch.
    [UIApplication sharedApplication].statusBarHidden = YES;
    
    NSString *path;
    path = [[NSBundle mainBundle] pathForResource:@"cat" ofType:@"jpg"];
    image = [[UIImage alloc] initWithContentsOfFile:path];
    
    imageview = [[UIImageView alloc] init];
    imageview.frame = CGRectMake(0, 0, 768, 576);
    imageview.image = image;
    
    [self.window setBackgroundColor:[UIColor blackColor]];
    [self.window addSubview:imageview];
   
    c_button = [UIButton buttonWithType:110];
    [c_button setTitle:@"Objective-C" forState:UIControlStateNormal];
    [c_button.titleLabel setFont:[UIFont systemFontOfSize:24]];
    c_button.frame = CGRectMake(8, 567+16, 245, 64);   
    [c_button addTarget:self action:@selector(c_buttonPressed) forControlEvents:UIControlEventTouchUpInside];
    [self.window addSubview:c_button];
 
    asm_button = [UIButton buttonWithType:110];
    [asm_button setTitle:@"ASM" forState:UIControlStateNormal];
    [asm_button.titleLabel setFont:[UIFont systemFontOfSize:24]];
    asm_button.frame = CGRectMake(8+245+8, 567+16, 245, 64);   
    [asm_button addTarget:self action:@selector(asm_buttonPressed) forControlEvents:UIControlEventTouchUpInside];
    [self.window addSubview:asm_button];    

    neon_button = [UIButton buttonWithType:110];
    [neon_button setTitle:@"NEON" forState:UIControlStateNormal];
    [neon_button.titleLabel setFont:[UIFont systemFontOfSize:24]];
    neon_button.frame = CGRectMake(8+245+8+245+8, 567+16, 245, 64);   
    [neon_button addTarget:self action:@selector(neon_buttonPressed) forControlEvents:UIControlEventTouchUpInside];
    [self.window addSubview:neon_button];   

    reset_button = [UIButton buttonWithType:110];
    [reset_button setTitle:@"reset" forState:UIControlStateNormal];
    [reset_button.titleLabel setFont:[UIFont systemFontOfSize:24]];
    reset_button.frame = CGRectMake(8, 567+16+64+8, 768-16, 64);   
    [reset_button addTarget:self action:@selector(reset_buttonPressed) forControlEvents:UIControlEventTouchUpInside];
    [self.window addSubview:reset_button]; 
    
    label = [[[UILabel alloc] initWithFrame:CGRectMake(0, 567+16+64+8+64+8, 768-16, 100)] autorelease];
    label.text = @"";
    label.backgroundColor = [UIColor blackColor];
    //label.textAlignment = UITextAlignmentCenter;
    label.textColor = [UIColor whiteColor];
    label.font = [UIFont systemFontOfSize:52]; 
    [self.window addSubview:label];  
    
    [self.window makeKeyAndVisible];
    return YES;
}

- (void)applicationWillResignActive:(UIApplication *)application
{
    /*
     Sent when the application is about to move from active to inactive state. This can occur for certain types of temporary interruptions (such as an incoming phone call or SMS message) or when the user quits the application and it begins the transition to the background state.
     Use this method to pause ongoing tasks, disable timers, and throttle down OpenGL ES frame rates. Games should use this method to pause the game.
     */
}

- (void)applicationDidEnterBackground:(UIApplication *)application
{
    /*
     Use this method to release shared resources, save user data, invalidate timers, and store enough application state information to restore your application to its current state in case it is terminated later. 
     If your application supports background execution, this method is called instead of applicationWillTerminate: when the user quits.
     */
}

- (void)applicationWillEnterForeground:(UIApplication *)application
{
    /*
     Called as part of the transition from the background to the inactive state; here you can undo many of the changes made on entering the background.
     */
}

- (void)applicationDidBecomeActive:(UIApplication *)application
{
    /*
     Restart any tasks that were paused (or not yet started) while the application was inactive. If the application was previously in the background, optionally refresh the user interface.
     */
}

- (void)applicationWillTerminate:(UIApplication *)application
{
    /*
     Called when the application is about to terminate.
     Save data if appropriate.
     See also applicationDidEnterBackground:.
     */
}

- (void)dealloc
{
    [image release];
    [mosaicized_image release];
    [imageview release];
    
    [_window release];
    [super dealloc];
}

@end
